from typing import Dict, List, Tuple, Optional
from collections import OrderedDict
import re
import math
from hein_robots.robotics import Location, Orientation, Cartesian
from hein_robots.grids import LocationGroup


class URScriptSequence(LocationGroup):
    LOCATION_REGEX = re.compile('global (\w+)_p=p\[(.*)\]')
    JOINT_REGEX = re.compile('global (\w+)_q=\[(.*)\]')

    @classmethod
    def parse_sequence_file(cls, file_path: str) -> Tuple[Dict[str, Location], Dict[str, List[float]]]:
        with open(file_path) as sequence_file:
            locations = OrderedDict()
            joints = OrderedDict()

            for line in sequence_file:
                location_match = cls.LOCATION_REGEX.search(line)
                if location_match:
                    name, location_str = location_match.groups()
                    location_list = [float(n) for n in location_str.split(', ')]
                    position = Cartesian(*location_list[:3])
                    rotation_vector = Cartesian(*location_list[3:])
                    locations[name] = Location(position, Orientation.from_rotation_vector(rotation_vector)).convert_m_to_mm()

                joint_match = cls.JOINT_REGEX.search(line)
                if joint_match:
                    name, joint_str = joint_match.groups()
                    joint_list = [math.degrees(float(n)) for n in joint_str.split(', ')]
                    joints[name] = joint_list

        return locations, joints

    def __init__(self, urscript_file: str, location_names: Optional[List[str]] = None):
        self.locations, self.joints = self.parse_sequence_file(urscript_file)

        if location_names is not None:
            for name in location_names:
                if name not in self.locations:
                    raise URScriptSequenceError(f'Location "{name}" not found in URScript file: {urscript_file}')

    def __len__(self):
        return len(self.locations)

    def __getitem__(self, item):
        return self.locations[item]

    def indexes(self) -> List[str]:
        return list(self.locations.keys())


class URScriptSequenceError(Exception):
    pass



from typing import Dict, List, Optional
import json
from hein_robots.robotics import Location
from hein_robots.grids import LocationGroup


class KinovaSequence(LocationGroup):
    @staticmethod
    def parse_sequence_file(sequence_file_path: str) -> Dict[str, Location]:
        with open(sequence_file_path) as sequence_file:
            locations = {}
            data = json.load(sequence_file)
            tasks = data['sequences']['sequence'][0]['tasks']

            for task in tasks:
                if 'reachPose' in task['action']:
                    pose = task['action']['reachPose']['targetPose']
                    location = Location(pose['x'], pose['y'], pose['z'], pose['thetaX'], pose['thetaY'], pose['thetaZ'])
                    locations[task['action']['name']] = location.convert_m_to_mm()

            return locations

    def __init__(self, sequence_json_file: str, location_names: Optional[List[str]] = None):
        self.locations = self.parse_sequence_file(sequence_json_file)

        if location_names is not None:
            for name in location_names:
                if name not in self.locations:
                    raise KinovaSequenceError(f'Location "{name}" not found in Kinova sequence file: {sequence_json_file}')

    def __len__(self):
        return len(self.locations)

    def __getitem__(self, item):
        return self.locations[item]

    def indexes(self) -> List[str]:
        return list(self.locations.keys())


class KinovaSequenceError(Exception):
    pass



from typing import Optional, List, Union
from hein_robots.robotics import Location, Twist, Wrench, Cartesian, Orientation, Units, Frame
from hein_robots.base.actuators import Actuator
from hein_robots.base.grippers import Gripper


class RobotArm:
    def __init__(self, default_velocity: float = 250, max_velocity: float = 500, position_units: str = Units.MILLIMETERS,
                 gripper_default_velocity: float = 0.5, gripper_default_force: float = 0.5):
        self._position_units = position_units
        self._default_velocity = default_velocity
        self._max_velocity = max_velocity
        self._gripper_default_velocity = gripper_default_velocity
        self._gripper_default_force = gripper_default_force

    @property
    def connected(self) -> bool:
        return False

    @property
    def max_velocity(self) -> float:
        return self._max_velocity

    @max_velocity.setter
    def max_velocity(self, value: float):
        self._max_velocity = value

    @property
    def default_velocity(self) -> float:
        return self._default_velocity

    @default_velocity.setter
    def default_velocity(self, value: float):
        if value < 0 or value > self.max_velocity:
            raise RobotArmInvalidVelocityError(f'Invalid velocity: {value} m/s, must be less than {self.max_velocity} m/s')

        self._default_velocity = value

    @property
    def acceleration(self) -> float:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} acceleration not supported')

    @property
    def velocity(self) -> float:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} velocity not supported')

    @property
    def location(self) -> Location:
        return Location()

    @property
    def twist(self) -> Twist:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} twist not supported')

    @property
    def wrench(self) -> Wrench:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} wrench not supported')

    @property
    def joint_positions(self) -> List[float]:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} joint_positions not supported')

    @property
    def joint_count(self) -> int:
        return len(self.joint_positions)

    @property
    def gripper_position(self) -> float:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} gripper_position not supported')

    @property
    def gripper_velocity(self) -> float:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} gripper_velocity not supported')

    @property
    def gripper_default_velocity(self) -> float:
        return self._gripper_default_velocity

    @gripper_default_velocity.setter
    def gripper_default_velocity(self, velocity: float):
        if velocity < 0 or velocity > 1:
            raise RobotArmGripperInvalidVelocityError(f'Invalid velocity, must be float between 0 and 1: {velocity}')

        self._gripper_default_velocity = velocity

    @property
    def gripper_default_force(self) -> float:
        return self._gripper_default_velocity

    @gripper_default_force.setter
    def gripper_default_force(self, force: float):
        if force < 0 or force > 1:
            raise RobotArmGripperInvalidForceError(f'Invalid force, must be float between 0 and 1: {force}')

        self._gripper_default_force = force

    @property
    def tool_offset(self) -> Location:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} tool_offset not supported')

    @tool_offset.setter
    def tool_offset(self, value: Location):
        raise RobotArmNotSupportedError(f'Setting {self.__class__.__name__} tool_offset not supported')

    @property
    def tool_mass(self) -> float:
        raise RobotArmNotSupportedError(f'Reading {self.__class__.__name__} tool_mass not supported')

    @tool_mass.setter
    def tool_mass(self, value: float):
        raise RobotArmNotSupportedError(f'Setting {self.__class__.__name__} tool_mass not supported')

    def connect(self):
        pass

    def disconnect(self):
        pass

    def stop(self):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} stop not supported')

    def pause(self):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} pause not supported')

    def resume(self):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} resume not supported')

    def emergency_stop(self):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} emergency_stop not supported')

    def clear_faults(self):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} clear_faults not supported')

    def home(self, wait: bool = True):
        pass

    def move(self, x: Optional[float] = None, y: Optional[float] = None, z: Optional[float] = None,
             rx: Optional[float] = None, ry: Optional[float] = None, rz: Optional[float] = None,
             frame: Optional[Frame] = None, velocity: Optional[float] = None, acceleration: Optional[float] = None,
             relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        if not relative:
            location = Location(
                x=self.location.position.x if x is None else x,
                y=self.location.position.y if y is None else y,
                z=self.location.position.z if z is None else z,
                ry=self.location.orientation.ry if ry is None else ry,
                rx=self.location.orientation.rx if rx is None else rx,
                rz=self.location.orientation.rz if rz is None else rz,
            )
        else:
            location = Location(x or 0, y or 0, z or 0, rx or 0, ry or 0, rz or 0)

        self.move_to_location(location, frame=frame, velocity=velocity, acceleration=acceleration, relative=relative, wait=wait,
                              timeout=timeout)

    def move_tool(self, x: Optional[float] = None, y: Optional[float] = None, z: Optional[float] = None,
             rx: Optional[float] = None, ry: Optional[float] = None, rz: Optional[float] = None,
             velocity: Optional[float] = None, acceleration: Optional[float] = None,
             wait: bool = True, timeout: Optional[float] = None):
        location = Location(x or 0, y or 0, z or 0, rx or 0, ry or 0, rz or 0)
        self.move_to_location(location, frame=self.location, velocity=velocity, acceleration=acceleration, relative=False, wait=wait,
                              timeout=timeout)

    def move_to_location(self, location: Location, frame: Optional[Frame] = None,
                         velocity: Optional[float] = None, acceleration: Optional[float] = None,
                         relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        pass

    def move_tool_to_location(self, location: Location,
                         velocity: Optional[float] = None, acceleration: Optional[float] = None,
                         relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        self.move_to_location(location, frame=self.location,
                              velocity=velocity, acceleration=acceleration,
                              relative=relative, wait=wait, timeout=timeout)

    def move_to_locations(self, *locations: Location, frame: Optional[Frame] = None,
                          velocity: Optional[float] = None, acceleration: Optional[float] = None,
                          relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        for location in locations:
            self.move_to_location(location, frame=frame,
                                  velocity=velocity, acceleration=acceleration,
                                  relative=relative, wait=wait, timeout=timeout)

    def move_tool_to_locations(self, *locations: Location,
                          velocity: Optional[float] = None, acceleration: Optional[float] = None,
                          relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        self.move_to_locations(*locations, velocity=velocity, acceleration=acceleration,
                               relative=relative, wait=wait, timeout=timeout)

    def move_joints(self, joint_positions: List[float],
                    velocity: Optional[float] = None, acceleration: Optional[float] = None,
                    relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        pass

    def move_joint(self, joint_id: int, position: float,
                   velocity: Optional[float] = None, acceleration: Optional[float] = None,
                   relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        positions = list(self.joint_positions)
        positions[joint_id] = positions[joint_id] + position if relative else position

        self.move_joints(positions, velocity=velocity, acceleration=acceleration, relative=relative, wait=wait, timeout=timeout)

    def move_twist(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, rx: float = 0.0, ry: float = 0.0, rz: float = 0.0,
                   duration: Optional[float] = None, wait: bool = True, timeout: Optional[float] = None):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} move_twist not supported')

    def move_twist_to(self, twist: Twist, duration: Optional[float] = None, wait: bool = True, timeout: Optional[float] = None):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} move_twist_to stop not supported')

    def wait(self, timeout: Optional[float] = None):
        pass

    def wait_for_gripper_stop(self, timeout: Optional[float] = None):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} gripper not supported')

    def open_gripper(self, position: Optional[Union[float, bool]] = None, wait: bool = True, timeout: Optional[float] = None):
        raise RobotArmNotSupportedError(f'{self.__class__.__name__} gripper not supported')

    def close_gripper(self, position: Optional[Union[float, bool]] = None, wait: bool = True, timeout: Optional[float] = None):
        if position is None:
            position = False
        elif isinstance(position, bool):
            position = not position

        self.open_gripper(position, wait=wait, timeout=timeout)


class RobotArmError(Exception):
    pass


class RobotArmNotConnectedError(RobotArmError):
    pass


class RobotArmConnectionError(RobotArmError):
    pass


class RobotArmNotSupportedError(RobotArmError):
    pass


class RobotArmUnitsError(RobotArmError):
    pass


class RobotArmMovementError(RobotArmError):
    pass


class RobotArmInvalidJointsError(RobotArmError):
    pass


class RobotArmInvalidVelocityError(RobotArmError):
    pass


class RobotArmWaitTimeoutError(RobotArmError):
    pass


class RobotArmGripperTimeoutError(RobotArmError):
    pass


class RobotArmGripperInvalidForceError(RobotArmError):
    pass


class RobotArmGripperInvalidVelocityError(RobotArmError):
    pass



from typing import Optional, List, Callable, Dict
import logging
import threading
import queue
import time
from kortex_api.TCPTransport import TCPTransport
from kortex_api.RouterClient import RouterClient, RouterClientSendOptions
from kortex_api.SessionManager import SessionManager
from kortex_api.autogen.messages import Session_pb2
from kortex_api.autogen.client_stubs.ControlConfigClientRpc import ControlConfigClient
from kortex_api.autogen.client_stubs.BaseClientRpc import BaseClient
from kortex_api.autogen.client_stubs.BaseCyclicClientRpc import BaseCyclicClient
from kortex_api.autogen.client_stubs.GripperCyclicClientRpc import GripperCyclicClient
from kortex_api.autogen.messages import Base_pb2, BaseCyclic_pb2, Common_pb2, ControlConfig_pb2

logger = logging.getLogger(__name__)


class KortexError(Exception):
    pass


class KortexActionError(KortexError):
    pass


class KortexActionAbortedError(KortexActionError):
    pass


class KortexActionTimeoutError(KortexActionError):
    pass


class KortexActionNotFoundError(KortexActionError):
    pass


class KortexConnection:
    def __init__(self, host: str = '192.168.1.10', port: int = 10000, username: str = 'admin', password: str = 'admin', connect: bool = True, action_timeout: float = 60.0):
        self.logger = logger.getChild(self.__class__.__name__)
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.connected = False
        self.action_timeout = action_timeout
        self.notification_listeners: Dict[int, List[Callable]] = {}

        self.transport = TCPTransport()
        self.router = RouterClient(self.transport, RouterClient.basicErrorCallback)
        self.session_manager: Optional[SessionManager] = None
        self.client: Optional[BaseClient] = None
        self.cyclic_client: Optional[BaseCyclicClient] = None
        self.control_client: Optional[ControlConfigClient] = None
        self.subscription: Optional[Base_pb2.NotificationHandle] = None

        if connect:
            self.connect()

    def connect(self):
        if self.connected:
            return

        self.transport.connect(self.host, self.port)

        session = Session_pb2.CreateSessionInfo()
        session.username = self.username
        session.password = self.password
        session.session_inactivity_timeout = 10000
        session.connection_inactivity_timeout = 2000

        self.session_manager = SessionManager(self.router)
        self.session_manager.CreateSession(session)
        self.client = BaseClient(self.router)
        self.cyclic_client = BaseCyclicClient(self.router)
        self.control_client = ControlConfigClient(self.router)
        self.subscription = self.client.OnNotificationActionTopic(self.dispatch_notification, Base_pb2.NotificationOptions())

        self.connected = True

    def disconnect(self):
        self.client.Unsubscribe(self.subscription)

    def dispatch_notification(self, notification: Base_pb2.ActionNotification):
        if notification.action_event in self.notification_listeners:
            for listener in self.notification_listeners[notification.action_event]:
                listener(notification)

    def add_notification_listener(self, listener: Callable, *action_types: int):
        for action_type in action_types:
            if action_type not in self.notification_listeners:
                self.notification_listeners[action_type] = []

            self.notification_listeners[action_type].append(listener)

    def remove_notification_listener(self, listener: Callable, *action_types: int):
        for action_type in action_types:
            if action_type not in self.notification_listeners or listener not in self.notification_listeners[action_type]:
                raise TypeError(f'Cannot remove listener, no matching listener with action_type "{action_type}" found')

            self.notification_listeners[action_type].remove(listener)

    def wait_for_notification(self, action_type: int, raise_on_abort: bool = True, timeout: Optional[float] = None) -> Base_pb2.ActionNotification:
        if timeout is None:
            timeout = self.action_timeout

        notification_queue = queue.Queue()
        action_types = [action_type, Base_pb2.ACTION_ABORT] if raise_on_abort else [action_type]
        self.add_notification_listener(lambda n: notification_queue.put(n), *action_types)

        try:
            notification = notification_queue.get(timeout=timeout)
        except queue.Empty:
            raise KortexActionTimeoutError(f'Kortex action timeout')

        if raise_on_abort and notification.action_event == Base_pb2.ACTION_ABORT:
            raise KortexActionAbortedError(f'Kortex action aborted: {Base_pb2.SubErrorCodes.Name(notification.abort_details)}')

        return notification

    def wait_for_action_end(self, timeout: Optional[float] = None, raise_on_abort: bool = True):
        self.wait_for_notification(Base_pb2.ACTION_END, raise_on_abort=raise_on_abort, timeout=timeout)

    def execute_action(self, action: Base_pb2.Action, wait: bool = True, timeout: Optional[float] = None):
        self.client.ExecuteAction(action)

        if wait:
            self.wait_for_action_end(timeout=timeout)

    def execute_gripper_command(self, command: Base_pb2.GripperCommand):
        self.client.SendGripperCommand(command)

    def execute_existing_action(self, name: str, action_type: int, raise_on_abort: bool = True, timeout: Optional[float] = None, wait: bool = True):
        requested_action_type = Base_pb2.RequestedActionType()
        requested_action_type.action_type = action_type
        action_list = self.client.ReadAllActions(requested_action_type)
        action_handle = None

        for action in action_list.action_list:
            if action.name == name:
                action_handle = action.handle

        if action_handle is None:
            raise KortexActionNotFoundError(f'Action not found: {name}')

        self.client.ExecuteActionFromReference(action_handle)

        if wait:
            self.wait_for_action_end(timeout=timeout, raise_on_abort=raise_on_abort)
